#include "pci_utils.h"
#include <stdlib.h>
#include <strings.h>

/* This file is based on work from pciutils and the Linux kernel. It is
   released under the GNU general public license version 2. */

extern struct pci_access *pacc;

struct device *
scan_device(struct pci_dev *p)
{
	int how_much = 256;
	struct device *d;

	d = malloc(sizeof(struct device));
	bzero(d, sizeof(*d));
	d->dev = p;
	if (!pci_read_block(p, 0, d->config, how_much)) {
		fprintf(stderr, "Unable to read %d bytes of configuration space.",
			how_much);
		return NULL;
	}

	d->config_cnt = how_much;
	pci_setup_cache(p, d->config, d->config_cnt);
	pci_fill_info(p, PCI_FILL_IDENT | PCI_FILL_IRQ | PCI_FILL_BASES | PCI_FILL_ROM_BASE | PCI_FILL_SIZES);
	return d;
}

static int
config_fetch(struct device *d, unsigned int pos, unsigned int len)
{
	if (pos + len < d->config_cnt)
		return 1;
	if (pacc->method != PCI_ACCESS_DUMP)
		return 0;
	return pci_read_block(d->dev, pos, d->config + pos, len);
}

/* Config space accesses */

static inline byte
get_conf_byte(struct device *d, unsigned int pos)
{
	return d->config[pos];
}

static word
get_conf_word(struct device *d, unsigned int pos)
{
	return d->config[pos] | (d->config[pos+1] << 8);
}

int
get_pm_cap(struct device *d)
{
	if (get_conf_word(d, PCI_STATUS) & PCI_STATUS_CAP_LIST)
	{
		int where = get_conf_byte(d, PCI_CAPABILITY_LIST) & ~3;
		while (where)
		{
			int id, next, cap;
			config_fetch(d, where, 4);
			id = get_conf_byte(d, where + PCI_CAP_LIST_ID);
			next = get_conf_byte(d, where + PCI_CAP_LIST_NEXT) & ~3;
			cap = get_conf_word(d, where + PCI_CAP_FLAGS);

			if (id == PCI_CAP_ID_PM) {
				return where;
			}
			where = next;
		}
	}
	return 0;
}

int
pci_set_power_state(struct pci_dev *dev, int state, int pm)
{
        u16 pmcsr;
	u16 pci_command;
	
        /* bound the state we're entering */
        if (state > 3) state = 3;

        /* check if this device supports the desired state */
        if (state == 1 || state == 2) {
                u16 pmc;
                pmc = pci_read_word(dev,pm + PCI_PM_PMC);
                if (state == 1 && !(pmc & PCI_PM_CAP_D1)) return -1;
                else if (state == 2 && !(pmc & PCI_PM_CAP_D2)) return -1;
        }

	if (state != 0) {
		/* Disable the device before powering it down */
		pci_command = pci_read_word(dev, PCI_COMMAND);
		if (pci_command & PCI_COMMAND_MASTER) {
			pci_command &= ~PCI_COMMAND_MASTER;
			pci_write_word(dev, PCI_COMMAND, pci_command);
		}
	}

	if (state == 0) {
		pmcsr = 0;
	} else {
		pmcsr = pci_read_word(dev, pm + PCI_PM_CTRL);
		pmcsr &= ~PCI_PM_CTRL_STATE_MASK;
		pmcsr |= state;
	}

        /* enter specified state */
        pci_write_word(dev, pm + PCI_PM_CTRL, pmcsr);

	return 0;
}
